mirror of
https://github.com/NVIDIA/cuda-samples.git
synced 2026-10-13 00:08:26 +08:00
CUDA 13.2 samples update (#432)
- Added Python samples for CUDA Python 1.0 release - Renamed top-level `Samples` directory to `cpp` to accommodate Python samples.
This commit is contained in:
18
cpp/3_CUDA_Features/jacobiCudaGraphs/.vscode/c_cpp_properties.json
vendored
Normal file
18
cpp/3_CUDA_Features/jacobiCudaGraphs/.vscode/c_cpp_properties.json
vendored
Normal file
@@ -0,0 +1,18 @@
|
||||
{
|
||||
"configurations": [
|
||||
{
|
||||
"name": "Linux",
|
||||
"includePath": [
|
||||
"${workspaceFolder}/**",
|
||||
"${workspaceFolder}/../../../Common"
|
||||
],
|
||||
"defines": [],
|
||||
"compilerPath": "/usr/local/cuda/bin/nvcc",
|
||||
"cStandard": "gnu17",
|
||||
"cppStandard": "gnu++14",
|
||||
"intelliSenseMode": "linux-gcc-x64",
|
||||
"configurationProvider": "ms-vscode.makefile-tools"
|
||||
}
|
||||
],
|
||||
"version": 4
|
||||
}
|
||||
7
cpp/3_CUDA_Features/jacobiCudaGraphs/.vscode/extensions.json
vendored
Normal file
7
cpp/3_CUDA_Features/jacobiCudaGraphs/.vscode/extensions.json
vendored
Normal file
@@ -0,0 +1,7 @@
|
||||
{
|
||||
"recommendations": [
|
||||
"nvidia.nsight-vscode-edition",
|
||||
"ms-vscode.cpptools",
|
||||
"ms-vscode.makefile-tools"
|
||||
]
|
||||
}
|
||||
38
cpp/3_CUDA_Features/jacobiCudaGraphs/CMakeLists.txt
Normal file
38
cpp/3_CUDA_Features/jacobiCudaGraphs/CMakeLists.txt
Normal file
@@ -0,0 +1,38 @@
|
||||
cmake_minimum_required(VERSION 3.20)
|
||||
|
||||
list(APPEND CMAKE_MODULE_PATH "${CMAKE_CURRENT_SOURCE_DIR}/../../../cmake/Modules")
|
||||
|
||||
project(jacobiCudaGraphs LANGUAGES C CXX CUDA)
|
||||
|
||||
find_package(CUDAToolkit REQUIRED)
|
||||
|
||||
set(CMAKE_POSITION_INDEPENDENT_CODE ON)
|
||||
|
||||
set(CMAKE_CUDA_ARCHITECTURES 75 80 86 87 89 90 100 110 120)
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -Wno-deprecated-gpu-targets")
|
||||
if(ENABLE_CUDA_DEBUG)
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -G") # enable cuda-gdb (may significantly affect performance on some targets)
|
||||
else()
|
||||
set(CMAKE_CUDA_FLAGS "${CMAKE_CUDA_FLAGS} -lineinfo") # add line information to all builds for debug tools (exclusive to -G option)
|
||||
endif()
|
||||
|
||||
# Include directories and libraries
|
||||
include_directories(../../../Common)
|
||||
|
||||
# Source file
|
||||
# Add target for jacobiCudaGraphs
|
||||
add_executable(jacobiCudaGraphs jacobi.cu main.cpp)
|
||||
|
||||
target_compile_options(jacobiCudaGraphs PRIVATE $<$<COMPILE_LANGUAGE:CUDA>:--extended-lambda>)
|
||||
|
||||
target_compile_features(jacobiCudaGraphs PRIVATE cxx_std_17 cuda_std_17)
|
||||
|
||||
set_target_properties(jacobiCudaGraphs PROPERTIES CUDA_SEPARABLE_COMPILATION ON)
|
||||
|
||||
target_include_directories(jacobiCudaGraphs PRIVATE
|
||||
${CUDAToolkit_INCLUDE_DIRS}
|
||||
)
|
||||
|
||||
# Include installation configuration
|
||||
include(${CMAKE_CURRENT_SOURCE_DIR}/../../../cmake/InstallSamples.cmake)
|
||||
setup_samples_install()
|
||||
30
cpp/3_CUDA_Features/jacobiCudaGraphs/README.md
Normal file
30
cpp/3_CUDA_Features/jacobiCudaGraphs/README.md
Normal file
@@ -0,0 +1,30 @@
|
||||
# jacobiCudaGraphs - Jacobi CUDA Graphs
|
||||
|
||||
## Description
|
||||
|
||||
Demonstrates Instantiated CUDA Graph Update with Jacobi Iterative Method using cudaGraphExecKernelNodeSetParams() and cudaGraphExecUpdate() approach.
|
||||
|
||||
## Key Concepts
|
||||
|
||||
CUDA Graphs, Stream Capture, Instantiated CUDA Graph Update, Cooperative Groups
|
||||
|
||||
## Supported SM Architectures
|
||||
|
||||
## Supported OSes
|
||||
|
||||
Linux, Windows
|
||||
|
||||
## Supported CPU Architecture
|
||||
|
||||
x86_64, armv7l
|
||||
|
||||
## CUDA APIs involved
|
||||
|
||||
### [CUDA Runtime API](http://docs.nvidia.com/cuda/cuda-runtime-api/index.html)
|
||||
cudaExtent, cudaGraphLaunch, cudaGraphAddMemcpyNode, cudaMallocHost, cudaPitchedPtr, cudaStreamEndCapture, cudaGraphCreate, cudaFreeHost, cudaMemsetAsync, cudaMemcpyAsync, cudaGraphExecKernelNodeSetParams, cudaStreamCreateWithFlags, cudaGraphInstantiate, cudaStreamBeginCapture, cudaFree, cudaGraphExecUpdate, cudaGraphAddKernelNode, cudaPos, cudaStreamSynchronize, cudaGraphAddMemsetNode, cudaMalloc, cudaGraphExecDestroy
|
||||
|
||||
## Prerequisites
|
||||
|
||||
Download and install the [CUDA Toolkit](https://developer.nvidia.com/cuda-downloads) for your corresponding platform.
|
||||
|
||||
## References (for more details)
|
||||
395
cpp/3_CUDA_Features/jacobiCudaGraphs/jacobi.cu
Normal file
395
cpp/3_CUDA_Features/jacobiCudaGraphs/jacobi.cu
Normal file
@@ -0,0 +1,395 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#include <cooperative_groups.h>
|
||||
#include <cuda_runtime.h>
|
||||
#include <helper_cuda.h>
|
||||
#include <vector>
|
||||
|
||||
#include "jacobi.h"
|
||||
|
||||
namespace cg = cooperative_groups;
|
||||
|
||||
// 8 Rows of square-matrix A processed by each CTA.
|
||||
// This can be max 32 and only power of 2 (i.e., 2/4/8/16/32).
|
||||
#define ROWS_PER_CTA 8
|
||||
|
||||
#if !defined(__CUDA_ARCH__) || __CUDA_ARCH__ >= 600
|
||||
#else
|
||||
__device__ double atomicAdd(double *address, double val)
|
||||
{
|
||||
unsigned long long int *address_as_ull = (unsigned long long int *)address;
|
||||
unsigned long long int old = *address_as_ull, assumed;
|
||||
|
||||
do {
|
||||
assumed = old;
|
||||
old = atomicCAS(address_as_ull, assumed, __double_as_longlong(val + __longlong_as_double(assumed)));
|
||||
|
||||
// Note: uses integer comparison to avoid hang in case of NaN (since NaN !=
|
||||
// NaN)
|
||||
} while (assumed != old);
|
||||
|
||||
return __longlong_as_double(old);
|
||||
}
|
||||
#endif
|
||||
|
||||
static __global__ void
|
||||
JacobiMethod(const float *A, const double *b, const float conv_threshold, double *x, double *x_new, double *sum)
|
||||
{
|
||||
// Handle to thread block group
|
||||
cg::thread_block cta = cg::this_thread_block();
|
||||
__shared__ double x_shared[N_ROWS]; // N_ROWS == n
|
||||
__shared__ double b_shared[ROWS_PER_CTA + 1];
|
||||
|
||||
for (int i = threadIdx.x; i < N_ROWS; i += blockDim.x) {
|
||||
x_shared[i] = x[i];
|
||||
}
|
||||
|
||||
if (threadIdx.x < ROWS_PER_CTA) {
|
||||
int k = threadIdx.x;
|
||||
for (int i = k + (blockIdx.x * ROWS_PER_CTA); (k < ROWS_PER_CTA) && (i < N_ROWS);
|
||||
k += ROWS_PER_CTA, i += ROWS_PER_CTA) {
|
||||
b_shared[i % (ROWS_PER_CTA + 1)] = b[i];
|
||||
}
|
||||
}
|
||||
|
||||
cg::sync(cta);
|
||||
|
||||
cg::thread_block_tile<32> tile32 = cg::tiled_partition<32>(cta);
|
||||
|
||||
for (int k = 0, i = blockIdx.x * ROWS_PER_CTA; (k < ROWS_PER_CTA) && (i < N_ROWS); k++, i++) {
|
||||
double rowThreadSum = 0.0;
|
||||
for (int j = threadIdx.x; j < N_ROWS; j += blockDim.x) {
|
||||
rowThreadSum += (A[i * N_ROWS + j] * x_shared[j]);
|
||||
}
|
||||
|
||||
for (int offset = tile32.size() / 2; offset > 0; offset /= 2) {
|
||||
rowThreadSum += tile32.shfl_down(rowThreadSum, offset);
|
||||
}
|
||||
|
||||
if (tile32.thread_rank() == 0) {
|
||||
atomicAdd(&b_shared[i % (ROWS_PER_CTA + 1)], -rowThreadSum);
|
||||
}
|
||||
}
|
||||
|
||||
cg::sync(cta);
|
||||
|
||||
if (threadIdx.x < ROWS_PER_CTA) {
|
||||
cg::thread_block_tile<ROWS_PER_CTA> tile8 = cg::tiled_partition<ROWS_PER_CTA>(cta);
|
||||
double temp_sum = 0.0;
|
||||
|
||||
int k = threadIdx.x;
|
||||
|
||||
for (int i = k + (blockIdx.x * ROWS_PER_CTA); (k < ROWS_PER_CTA) && (i < N_ROWS);
|
||||
k += ROWS_PER_CTA, i += ROWS_PER_CTA) {
|
||||
double dx = b_shared[i % (ROWS_PER_CTA + 1)];
|
||||
dx /= A[i * N_ROWS + i];
|
||||
|
||||
x_new[i] = (x_shared[i] + dx);
|
||||
temp_sum += fabs(dx);
|
||||
}
|
||||
|
||||
for (int offset = tile8.size() / 2; offset > 0; offset /= 2) {
|
||||
temp_sum += tile8.shfl_down(temp_sum, offset);
|
||||
}
|
||||
|
||||
if (tile8.thread_rank() == 0) {
|
||||
atomicAdd(sum, temp_sum);
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
// Thread block size for finalError kernel should be multiple of 32
|
||||
static __global__ void finalError(double *x, double *g_sum)
|
||||
{
|
||||
// Handle to thread block group
|
||||
cg::thread_block cta = cg::this_thread_block();
|
||||
extern __shared__ double warpSum[];
|
||||
double sum = 0.0;
|
||||
|
||||
int globalThreadId = blockIdx.x * blockDim.x + threadIdx.x;
|
||||
|
||||
for (int i = globalThreadId; i < N_ROWS; i += blockDim.x * gridDim.x) {
|
||||
double d = x[i] - 1.0;
|
||||
sum += fabs(d);
|
||||
}
|
||||
|
||||
cg::thread_block_tile<32> tile32 = cg::tiled_partition<32>(cta);
|
||||
|
||||
for (int offset = tile32.size() / 2; offset > 0; offset /= 2) {
|
||||
sum += tile32.shfl_down(sum, offset);
|
||||
}
|
||||
|
||||
if (tile32.thread_rank() == 0) {
|
||||
warpSum[threadIdx.x / warpSize] = sum;
|
||||
}
|
||||
|
||||
cg::sync(cta);
|
||||
|
||||
double blockSum = 0.0;
|
||||
if (threadIdx.x < (blockDim.x / warpSize)) {
|
||||
blockSum = warpSum[threadIdx.x];
|
||||
}
|
||||
|
||||
if (threadIdx.x < 32) {
|
||||
for (int offset = tile32.size() / 2; offset > 0; offset /= 2) {
|
||||
blockSum += tile32.shfl_down(blockSum, offset);
|
||||
}
|
||||
if (tile32.thread_rank() == 0) {
|
||||
atomicAdd(g_sum, blockSum);
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
double JacobiMethodGpuCudaGraphExecKernelSetParams(const float *A,
|
||||
const double *b,
|
||||
const float conv_threshold,
|
||||
const int max_iter,
|
||||
double *x,
|
||||
double *x_new,
|
||||
cudaStream_t stream)
|
||||
{
|
||||
// CTA size
|
||||
dim3 nthreads(256, 1, 1);
|
||||
// grid size
|
||||
dim3 nblocks((N_ROWS / ROWS_PER_CTA) + 2, 1, 1);
|
||||
cudaGraph_t graph;
|
||||
cudaGraphExec_t graphExec = NULL;
|
||||
|
||||
double sum = 0.0;
|
||||
double *d_sum = NULL;
|
||||
checkCudaErrors(cudaMalloc(&d_sum, sizeof(double)));
|
||||
|
||||
std::vector<cudaGraphNode_t> nodeDependencies;
|
||||
cudaGraphNode_t memcpyNode, jacobiKernelNode, memsetNode;
|
||||
cudaMemcpy3DParms memcpyParams = {0};
|
||||
cudaMemsetParams memsetParams = {0};
|
||||
|
||||
memsetParams.dst = (void *)d_sum;
|
||||
memsetParams.value = 0;
|
||||
memsetParams.pitch = 0;
|
||||
// elementSize can be max 4 bytes, so we take sizeof(float) and width=2
|
||||
memsetParams.elementSize = sizeof(float);
|
||||
memsetParams.width = 2;
|
||||
memsetParams.height = 1;
|
||||
|
||||
checkCudaErrors(cudaGraphCreate(&graph, 0));
|
||||
checkCudaErrors(cudaGraphAddMemsetNode(&memsetNode, graph, NULL, 0, &memsetParams));
|
||||
nodeDependencies.push_back(memsetNode);
|
||||
|
||||
cudaKernelNodeParams NodeParams0, NodeParams1;
|
||||
NodeParams0.func = (void *)JacobiMethod;
|
||||
NodeParams0.gridDim = nblocks;
|
||||
NodeParams0.blockDim = nthreads;
|
||||
NodeParams0.sharedMemBytes = 0;
|
||||
void *kernelArgs0[6] = {
|
||||
(void *)&A, (void *)&b, (void *)&conv_threshold, (void *)&x, (void *)&x_new, (void *)&d_sum};
|
||||
NodeParams0.kernelParams = kernelArgs0;
|
||||
NodeParams0.extra = NULL;
|
||||
|
||||
checkCudaErrors(cudaGraphAddKernelNode(
|
||||
&jacobiKernelNode, graph, nodeDependencies.data(), nodeDependencies.size(), &NodeParams0));
|
||||
|
||||
nodeDependencies.clear();
|
||||
nodeDependencies.push_back(jacobiKernelNode);
|
||||
|
||||
memcpyParams.srcArray = NULL;
|
||||
memcpyParams.srcPos = make_cudaPos(0, 0, 0);
|
||||
memcpyParams.srcPtr = make_cudaPitchedPtr(d_sum, sizeof(double), 1, 1);
|
||||
memcpyParams.dstArray = NULL;
|
||||
memcpyParams.dstPos = make_cudaPos(0, 0, 0);
|
||||
memcpyParams.dstPtr = make_cudaPitchedPtr(&sum, sizeof(double), 1, 1);
|
||||
memcpyParams.extent = make_cudaExtent(sizeof(double), 1, 1);
|
||||
memcpyParams.kind = cudaMemcpyDeviceToHost;
|
||||
|
||||
checkCudaErrors(
|
||||
cudaGraphAddMemcpyNode(&memcpyNode, graph, nodeDependencies.data(), nodeDependencies.size(), &memcpyParams));
|
||||
|
||||
checkCudaErrors(cudaGraphInstantiate(&graphExec, graph, NULL, NULL, 0));
|
||||
|
||||
NodeParams1.func = (void *)JacobiMethod;
|
||||
NodeParams1.gridDim = nblocks;
|
||||
NodeParams1.blockDim = nthreads;
|
||||
NodeParams1.sharedMemBytes = 0;
|
||||
void *kernelArgs1[6] = {
|
||||
(void *)&A, (void *)&b, (void *)&conv_threshold, (void *)&x_new, (void *)&x, (void *)&d_sum};
|
||||
NodeParams1.kernelParams = kernelArgs1;
|
||||
NodeParams1.extra = NULL;
|
||||
|
||||
int k = 0;
|
||||
for (k = 0; k < max_iter; k++) {
|
||||
checkCudaErrors(cudaGraphExecKernelNodeSetParams(
|
||||
graphExec, jacobiKernelNode, ((k & 1) == 0) ? &NodeParams0 : &NodeParams1));
|
||||
checkCudaErrors(cudaGraphLaunch(graphExec, stream));
|
||||
checkCudaErrors(cudaStreamSynchronize(stream));
|
||||
|
||||
if (sum <= conv_threshold) {
|
||||
checkCudaErrors(cudaMemsetAsync(d_sum, 0, sizeof(double), stream));
|
||||
nblocks.x = (N_ROWS / nthreads.x) + 1;
|
||||
size_t sharedMemSize = ((nthreads.x / 32) + 1) * sizeof(double);
|
||||
if ((k & 1) == 0) {
|
||||
finalError<<<nblocks, nthreads, sharedMemSize, stream>>>(x_new, d_sum);
|
||||
}
|
||||
else {
|
||||
finalError<<<nblocks, nthreads, sharedMemSize, stream>>>(x, d_sum);
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaMemcpyAsync(&sum, d_sum, sizeof(double), cudaMemcpyDeviceToHost, stream));
|
||||
checkCudaErrors(cudaStreamSynchronize(stream));
|
||||
printf("GPU iterations : %d\n", k + 1);
|
||||
printf("GPU error : %.3e\n", sum);
|
||||
break;
|
||||
}
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaFree(d_sum));
|
||||
return sum;
|
||||
}
|
||||
|
||||
double JacobiMethodGpuCudaGraphExecUpdate(const float *A,
|
||||
const double *b,
|
||||
const float conv_threshold,
|
||||
const int max_iter,
|
||||
double *x,
|
||||
double *x_new,
|
||||
cudaStream_t stream)
|
||||
{
|
||||
// CTA size
|
||||
dim3 nthreads(256, 1, 1);
|
||||
// grid size
|
||||
dim3 nblocks((N_ROWS / ROWS_PER_CTA) + 2, 1, 1);
|
||||
cudaGraph_t graph;
|
||||
cudaGraphExec_t graphExec = NULL;
|
||||
|
||||
double sum = 0.0;
|
||||
double *d_sum;
|
||||
checkCudaErrors(cudaMalloc(&d_sum, sizeof(double)));
|
||||
|
||||
int k = 0;
|
||||
for (k = 0; k < max_iter; k++) {
|
||||
checkCudaErrors(cudaStreamBeginCapture(stream, cudaStreamCaptureModeGlobal));
|
||||
checkCudaErrors(cudaMemsetAsync(d_sum, 0, sizeof(double), stream));
|
||||
if ((k & 1) == 0) {
|
||||
JacobiMethod<<<nblocks, nthreads, 0, stream>>>(A, b, conv_threshold, x, x_new, d_sum);
|
||||
}
|
||||
else {
|
||||
JacobiMethod<<<nblocks, nthreads, 0, stream>>>(A, b, conv_threshold, x_new, x, d_sum);
|
||||
}
|
||||
checkCudaErrors(cudaMemcpyAsync(&sum, d_sum, sizeof(double), cudaMemcpyDeviceToHost, stream));
|
||||
checkCudaErrors(cudaStreamEndCapture(stream, &graph));
|
||||
|
||||
if (graphExec == NULL) {
|
||||
checkCudaErrors(cudaGraphInstantiate(&graphExec, graph, NULL, NULL, 0));
|
||||
}
|
||||
else {
|
||||
cudaGraphExecUpdateResult updateResult_out;
|
||||
checkCudaErrors(cudaGraphExecUpdate(graphExec, graph, NULL, &updateResult_out));
|
||||
if (updateResult_out != cudaGraphExecUpdateSuccess) {
|
||||
if (graphExec != NULL) {
|
||||
checkCudaErrors(cudaGraphExecDestroy(graphExec));
|
||||
}
|
||||
printf("k = %d graph update failed with error - %d\n", k, updateResult_out);
|
||||
checkCudaErrors(cudaGraphInstantiate(&graphExec, graph, NULL, NULL, 0));
|
||||
}
|
||||
}
|
||||
checkCudaErrors(cudaGraphLaunch(graphExec, stream));
|
||||
checkCudaErrors(cudaStreamSynchronize(stream));
|
||||
|
||||
if (sum <= conv_threshold) {
|
||||
checkCudaErrors(cudaMemsetAsync(d_sum, 0, sizeof(double), stream));
|
||||
nblocks.x = (N_ROWS / nthreads.x) + 1;
|
||||
size_t sharedMemSize = ((nthreads.x / 32) + 1) * sizeof(double);
|
||||
if ((k & 1) == 0) {
|
||||
finalError<<<nblocks, nthreads, sharedMemSize, stream>>>(x_new, d_sum);
|
||||
}
|
||||
else {
|
||||
finalError<<<nblocks, nthreads, sharedMemSize, stream>>>(x, d_sum);
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaMemcpyAsync(&sum, d_sum, sizeof(double), cudaMemcpyDeviceToHost, stream));
|
||||
checkCudaErrors(cudaStreamSynchronize(stream));
|
||||
printf("GPU iterations : %d\n", k + 1);
|
||||
printf("GPU error : %.3e\n", sum);
|
||||
break;
|
||||
}
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaFree(d_sum));
|
||||
return sum;
|
||||
}
|
||||
|
||||
double JacobiMethodGpu(const float *A,
|
||||
const double *b,
|
||||
const float conv_threshold,
|
||||
const int max_iter,
|
||||
double *x,
|
||||
double *x_new,
|
||||
cudaStream_t stream)
|
||||
{
|
||||
// CTA size
|
||||
dim3 nthreads(256, 1, 1);
|
||||
// grid size
|
||||
dim3 nblocks((N_ROWS / ROWS_PER_CTA) + 2, 1, 1);
|
||||
|
||||
double sum = 0.0;
|
||||
double *d_sum;
|
||||
checkCudaErrors(cudaMalloc(&d_sum, sizeof(double)));
|
||||
int k = 0;
|
||||
|
||||
for (k = 0; k < max_iter; k++) {
|
||||
checkCudaErrors(cudaMemsetAsync(d_sum, 0, sizeof(double), stream));
|
||||
if ((k & 1) == 0) {
|
||||
JacobiMethod<<<nblocks, nthreads, 0, stream>>>(A, b, conv_threshold, x, x_new, d_sum);
|
||||
}
|
||||
else {
|
||||
JacobiMethod<<<nblocks, nthreads, 0, stream>>>(A, b, conv_threshold, x_new, x, d_sum);
|
||||
}
|
||||
checkCudaErrors(cudaMemcpyAsync(&sum, d_sum, sizeof(double), cudaMemcpyDeviceToHost, stream));
|
||||
checkCudaErrors(cudaStreamSynchronize(stream));
|
||||
|
||||
if (sum <= conv_threshold) {
|
||||
checkCudaErrors(cudaMemsetAsync(d_sum, 0, sizeof(double), stream));
|
||||
nblocks.x = (N_ROWS / nthreads.x) + 1;
|
||||
size_t sharedMemSize = ((nthreads.x / 32) + 1) * sizeof(double);
|
||||
if ((k & 1) == 0) {
|
||||
finalError<<<nblocks, nthreads, sharedMemSize, stream>>>(x_new, d_sum);
|
||||
}
|
||||
else {
|
||||
finalError<<<nblocks, nthreads, sharedMemSize, stream>>>(x, d_sum);
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaMemcpyAsync(&sum, d_sum, sizeof(double), cudaMemcpyDeviceToHost, stream));
|
||||
checkCudaErrors(cudaStreamSynchronize(stream));
|
||||
printf("GPU iterations : %d\n", k + 1);
|
||||
printf("GPU error : %.3e\n", sum);
|
||||
break;
|
||||
}
|
||||
}
|
||||
|
||||
checkCudaErrors(cudaFree(d_sum));
|
||||
return sum;
|
||||
}
|
||||
33
cpp/3_CUDA_Features/jacobiCudaGraphs/jacobi.h
Normal file
33
cpp/3_CUDA_Features/jacobiCudaGraphs/jacobi.h
Normal file
@@ -0,0 +1,33 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
#ifndef JACOBI_H
|
||||
#define JACOBI_H
|
||||
|
||||
#define N_ROWS 512
|
||||
|
||||
#endif
|
||||
226
cpp/3_CUDA_Features/jacobiCudaGraphs/main.cpp
Normal file
226
cpp/3_CUDA_Features/jacobiCudaGraphs/main.cpp
Normal file
@@ -0,0 +1,226 @@
|
||||
/* Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions
|
||||
* are met:
|
||||
* * Redistributions of source code must retain the above copyright
|
||||
* notice, this list of conditions and the following disclaimer.
|
||||
* * Redistributions in binary form must reproduce the above copyright
|
||||
* notice, this list of conditions and the following disclaimer in the
|
||||
* documentation and/or other materials provided with the distribution.
|
||||
* * Neither the name of NVIDIA CORPORATION nor the names of its
|
||||
* contributors may be used to endorse or promote products derived
|
||||
* from this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
|
||||
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
|
||||
* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
|
||||
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
|
||||
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
|
||||
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
|
||||
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
|
||||
* OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
|
||||
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*/
|
||||
|
||||
// This sample demonstrates Instantiated CUDA Graph Update
|
||||
// with Jacobi Iterative Method in 3 different methods:
|
||||
// 1 - JacobiMethodGpuCudaGraphExecKernelSetParams() - CUDA Graph with
|
||||
// cudaGraphExecKernelNodeSetParams() 2 - JacobiMethodGpuCudaGraphExecUpdate() -
|
||||
// CUDA Graph with cudaGraphExecUpdate() 3 - JacobiMethodGpu() - Non CUDA Graph
|
||||
// method
|
||||
|
||||
// Jacobi method on a linear system A*x = b,
|
||||
// where A is diagonally dominant and the exact solution consists
|
||||
// of all ones.
|
||||
// The dimension N_ROWS is included in jacobi.h
|
||||
|
||||
#include <cuda_runtime.h>
|
||||
#include <helper_cuda.h>
|
||||
#include <helper_timer.h>
|
||||
#include <math.h>
|
||||
#include <stdio.h>
|
||||
#include <stdlib.h>
|
||||
|
||||
#include "jacobi.h"
|
||||
|
||||
// Run the Jacobi method for A*x = b on GPU with CUDA Graph -
|
||||
// cudaGraphExecKernelNodeSetParams().
|
||||
extern double JacobiMethodGpuCudaGraphExecKernelSetParams(const float *A,
|
||||
const double *b,
|
||||
const float conv_threshold,
|
||||
const int max_iter,
|
||||
double *x,
|
||||
double *x_new,
|
||||
cudaStream_t stream);
|
||||
|
||||
// Run the Jacobi method for A*x = b on GPU with Instantiated CUDA Graph Update
|
||||
// API - cudaGraphExecUpdate().
|
||||
extern double JacobiMethodGpuCudaGraphExecUpdate(const float *A,
|
||||
const double *b,
|
||||
const float conv_threshold,
|
||||
const int max_iter,
|
||||
double *x,
|
||||
double *x_new,
|
||||
cudaStream_t stream);
|
||||
|
||||
// Run the Jacobi method for A*x = b on GPU without CUDA Graph.
|
||||
extern double JacobiMethodGpu(const float *A,
|
||||
const double *b,
|
||||
const float conv_threshold,
|
||||
const int max_iter,
|
||||
double *x,
|
||||
double *x_new,
|
||||
cudaStream_t stream);
|
||||
|
||||
// creates N_ROWS x N_ROWS matrix A with N_ROWS+1 on the diagonal and 1
|
||||
// elsewhere. The elements of the right hand side b all equal 2*n, hence the
|
||||
// exact solution x to A*x = b is a vector of ones.
|
||||
void createLinearSystem(float *A, double *b);
|
||||
|
||||
// Run the Jacobi method for A*x = b on CPU.
|
||||
void JacobiMethodCPU(float *A, double *b, float conv_threshold, int max_iter, int *numit, double *x);
|
||||
|
||||
int main(int argc, char **argv)
|
||||
{
|
||||
if (checkCmdLineFlag(argc, (const char **)argv, "help")) {
|
||||
printf("Command line: jacobiCudaGraphs [-option]\n");
|
||||
printf("Valid options:\n");
|
||||
printf("-gpumethod=<0,1 or 2> : 0 - [Default] "
|
||||
"JacobiMethodGpuCudaGraphExecKernelSetParams\n");
|
||||
printf(" : 1 - JacobiMethodGpuCudaGraphExecUpdate\n");
|
||||
printf(" : 2 - JacobiMethodGpu - Non CUDA Graph\n");
|
||||
printf("-device=device_num : cuda device id");
|
||||
printf("-help : Output a help message\n");
|
||||
exit(EXIT_SUCCESS);
|
||||
}
|
||||
|
||||
int gpumethod = 0;
|
||||
if (checkCmdLineFlag(argc, (const char **)argv, "gpumethod")) {
|
||||
gpumethod = getCmdLineArgumentInt(argc, (const char **)argv, "gpumethod");
|
||||
|
||||
if (gpumethod < 0 || gpumethod > 2) {
|
||||
printf("Error: gpumethod must be 0 or 1 or 2, gpumethod=%d is invalid\n", gpumethod);
|
||||
exit(EXIT_SUCCESS);
|
||||
}
|
||||
}
|
||||
|
||||
int dev = findCudaDevice(argc, (const char **)argv);
|
||||
|
||||
double *b = NULL;
|
||||
float *A = NULL;
|
||||
checkCudaErrors(cudaMallocHost(&b, N_ROWS * sizeof(double)));
|
||||
memset(b, 0, N_ROWS * sizeof(double));
|
||||
checkCudaErrors(cudaMallocHost(&A, N_ROWS * N_ROWS * sizeof(float)));
|
||||
memset(A, 0, N_ROWS * N_ROWS * sizeof(float));
|
||||
|
||||
createLinearSystem(A, b);
|
||||
double *x = NULL;
|
||||
// start with array of all zeroes
|
||||
x = (double *)calloc(N_ROWS, sizeof(double));
|
||||
|
||||
float conv_threshold = 1.0e-2;
|
||||
int max_iter = 4 * N_ROWS * N_ROWS;
|
||||
int cnt = 0;
|
||||
|
||||
// create timer
|
||||
StopWatchInterface *timerCPU = NULL, *timerGpu = NULL;
|
||||
sdkCreateTimer(&timerCPU);
|
||||
|
||||
sdkStartTimer(&timerCPU);
|
||||
JacobiMethodCPU(A, b, conv_threshold, max_iter, &cnt, x);
|
||||
|
||||
double sum = 0.0;
|
||||
// Compute error
|
||||
for (int i = 0; i < N_ROWS; i++) {
|
||||
double d = x[i] - 1.0;
|
||||
sum += fabs(d);
|
||||
}
|
||||
sdkStopTimer(&timerCPU);
|
||||
printf("CPU iterations : %d\n", cnt);
|
||||
printf("CPU error : %.3e\n", sum);
|
||||
printf("CPU Processing time: %f (ms)\n", sdkGetTimerValue(&timerCPU));
|
||||
|
||||
float *d_A;
|
||||
double *d_b, *d_x, *d_x_new;
|
||||
cudaStream_t stream1;
|
||||
checkCudaErrors(cudaStreamCreateWithFlags(&stream1, cudaStreamNonBlocking));
|
||||
checkCudaErrors(cudaMalloc(&d_b, sizeof(double) * N_ROWS));
|
||||
checkCudaErrors(cudaMalloc(&d_A, sizeof(float) * N_ROWS * N_ROWS));
|
||||
checkCudaErrors(cudaMalloc(&d_x, sizeof(double) * N_ROWS));
|
||||
checkCudaErrors(cudaMalloc(&d_x_new, sizeof(double) * N_ROWS));
|
||||
|
||||
checkCudaErrors(cudaMemsetAsync(d_x, 0, sizeof(double) * N_ROWS, stream1));
|
||||
checkCudaErrors(cudaMemsetAsync(d_x_new, 0, sizeof(double) * N_ROWS, stream1));
|
||||
checkCudaErrors(cudaMemcpyAsync(d_A, A, sizeof(float) * N_ROWS * N_ROWS, cudaMemcpyHostToDevice, stream1));
|
||||
checkCudaErrors(cudaMemcpyAsync(d_b, b, sizeof(double) * N_ROWS, cudaMemcpyHostToDevice, stream1));
|
||||
|
||||
sdkCreateTimer(&timerGpu);
|
||||
sdkStartTimer(&timerGpu);
|
||||
|
||||
double sumGPU = 0.0;
|
||||
if (gpumethod == 0) {
|
||||
sumGPU = JacobiMethodGpuCudaGraphExecKernelSetParams(d_A, d_b, conv_threshold, max_iter, d_x, d_x_new, stream1);
|
||||
}
|
||||
else if (gpumethod == 1) {
|
||||
sumGPU = JacobiMethodGpuCudaGraphExecUpdate(d_A, d_b, conv_threshold, max_iter, d_x, d_x_new, stream1);
|
||||
}
|
||||
else if (gpumethod == 2) {
|
||||
sumGPU = JacobiMethodGpu(d_A, d_b, conv_threshold, max_iter, d_x, d_x_new, stream1);
|
||||
}
|
||||
|
||||
sdkStopTimer(&timerGpu);
|
||||
printf("GPU Processing time: %f (ms)\n", sdkGetTimerValue(&timerGpu));
|
||||
|
||||
checkCudaErrors(cudaFree(d_b));
|
||||
checkCudaErrors(cudaFree(d_A));
|
||||
checkCudaErrors(cudaFree(d_x));
|
||||
checkCudaErrors(cudaFree(d_x_new));
|
||||
|
||||
checkCudaErrors(cudaFreeHost(A));
|
||||
checkCudaErrors(cudaFreeHost(b));
|
||||
|
||||
printf("&&&& jacobiCudaGraphs %s\n", (fabs(sum - sumGPU) < conv_threshold) ? "PASSED" : "FAILED");
|
||||
|
||||
return (fabs(sum - sumGPU) < conv_threshold) ? EXIT_SUCCESS : EXIT_FAILURE;
|
||||
}
|
||||
|
||||
void createLinearSystem(float *A, double *b)
|
||||
{
|
||||
int i, j;
|
||||
for (i = 0; i < N_ROWS; i++) {
|
||||
b[i] = 2.0 * N_ROWS;
|
||||
for (j = 0; j < N_ROWS; j++)
|
||||
A[i * N_ROWS + j] = 1.0;
|
||||
A[i * N_ROWS + i] = N_ROWS + 1.0;
|
||||
}
|
||||
}
|
||||
|
||||
void JacobiMethodCPU(float *A, double *b, float conv_threshold, int max_iter, int *num_iter, double *x)
|
||||
{
|
||||
double *x_new;
|
||||
x_new = (double *)calloc(N_ROWS, sizeof(double));
|
||||
int k;
|
||||
|
||||
for (k = 0; k < max_iter; k++) {
|
||||
double sum = 0.0;
|
||||
for (int i = 0; i < N_ROWS; i++) {
|
||||
double temp_dx = b[i];
|
||||
for (int j = 0; j < N_ROWS; j++)
|
||||
temp_dx -= A[i * N_ROWS + j] * x[j];
|
||||
temp_dx /= A[i * N_ROWS + i];
|
||||
x_new[i] += temp_dx;
|
||||
sum += fabs(temp_dx);
|
||||
}
|
||||
|
||||
for (int i = 0; i < N_ROWS; i++)
|
||||
x[i] = x_new[i];
|
||||
|
||||
if (sum <= conv_threshold)
|
||||
break;
|
||||
}
|
||||
*num_iter = k + 1;
|
||||
free(x_new);
|
||||
}
|
||||
Reference in New Issue
Block a user